from contextlib import contextmanager
from typing import Any, Awaitable, Callable, List, Optional

import pandas as pd
from PIL import Image


class Subject:
    """Base class for implementing the observer pattern."""

    def __init__(self):
        self.observers: List[Callable[[], None]] = []

    def attach(self, observer: Callable[[], None]):
        """Attach an observer to the subject."""
        self.observers.append(observer)

    def detach(self, observer: Callable[[], None]):
        """Detach an observer from the subject."""
        self.observers.remove(observer)

    def notify(self):
        """Notify all observers about an event."""
        for observer in self.observers:
            if observer is None:
                continue
            observer()


class AsyncSubject:
    """Base class for implementing the async observer pattern."""

    def __init__(self):
        self.observers: List[Callable[[], Awaitable]] = []

    def attach(self, observer: Callable[[], Awaitable]):
        """Attach an observer to the subject."""
        self.observers.append(observer)

    def detach(self, observer: Callable[[], Awaitable]):
        """Detach an observer from the subject."""
        self.observers.remove(observer)

    async def notify(self):
        """Notify all observers about an event."""
        for observer in self.observers:
            if observer is None:
                continue
            await observer()


class CacheManager(Subject):
    """Manages cache for different clients and notifies observers on changes."""

    def __init__(self):
        super().__init__()
        self.CACHE = {}
        self.current_client_id = None
        self.current_cache = {}

    @contextmanager
    def set_client_id(self, client_id: str):
        """
        Context manager to set the current client_id and associated cache.

        Args:
            client_id (str): The client identifier.
        """
        previous_client_id = self.current_client_id
        self.current_client_id = client_id
        self.current_cache = self.CACHE.setdefault(client_id, {})
        try:
            yield
        finally:
            self.current_client_id = previous_client_id
            self.current_cache = self.CACHE.get(self.current_client_id, {})

    def add(self, name: str, obj: Any, obj_type: str, extension: Optional[str] = None):
        """
        Add an object to the current client's cache.

        Args:
            name (str): The cache key.
            obj (Any): The object to cache.
            obj_type (str): The type of the object.
        """
        object_extensions = {
            "image": "png",
            "pandas": "csv",
        }
        if obj_type in object_extensions:
            _extension = object_extensions[obj_type]
        else:
            _extension = type(obj).__name__.lower()
        self.current_cache[name] = {
            "obj": obj,
            "type": obj_type,
            "extension": extension or _extension,
        }
        self.notify()

    def add_pandas(self, name: str, obj: Any):
        """
        Add a pandas DataFrame or Series to the current client's cache.

        Args:
            name (str): The cache key.
            obj (Any): The pandas DataFrame or Series object.
        """
        if isinstance(obj, (pd.DataFrame, pd.Series)):
            self.add(name, obj.to_csv(), "pandas", extension="csv")
        else:
            raise ValueError("Object is not a pandas DataFrame or Series")

    def add_image(self, name: str, obj: Any, extension: str = "png"):
        """
        Add a PIL Image to the current client's cache.

        Args:
            name (str): The cache key.
            obj (Any): The PIL Image object.
        """
        if isinstance(obj, Image.Image):
            self.add(name, obj, "image", extension=extension)
        else:
            raise ValueError("Object is not a PIL Image")

    def get(self, name: str):
        """
        Get an object from the current client's cache.

        Args:
            name (str): The cache key.

        Returns:
            The cached object associated with the given cache key.
        """
        return self.current_cache[name]

    def get_last(self):
        """
        Get the last added item in the current client's cache.

        Returns:
            The last added item in the cache.
        """
        return list(self.current_cache.values())[-1]


cache_manager = CacheManager()



from typing import Dict, List, Optional

from langflow.interface.base import LangChainTypeCreator
from langflow.interface.custom_lists import documentloaders_type_to_cls_dict
from langflow.settings import settings
from langflow.utils.logger import logger
from langflow.utils.util import build_template_from_class


def build_file_path_template(
    suffixes: list, fileTypes: list, name: str = "file_path"
) -> Dict:
    """Build a file path template for a document loader."""
    return {
        "type": "file",
        "required": True,
        "show": True,
        "name": name,
        "value": "",
        "suffixes": suffixes,
        "fileTypes": fileTypes,
    }


class DocumentLoaderCreator(LangChainTypeCreator):
    type_name: str = "documentloaders"

    @property
    def type_to_loader_dict(self) -> Dict:
        return documentloaders_type_to_cls_dict

    def get_signature(self, name: str) -> Optional[Dict]:
        """Get the signature of a document loader."""
        try:
            signature = build_template_from_class(
                name, documentloaders_type_to_cls_dict
            )

            file_path_templates = {
                "AirbyteJSONLoader": build_file_path_template(
                    suffixes=[".json"], fileTypes=["json"]
                ),
                "CoNLLULoader": build_file_path_template(
                    suffixes=[".csv"], fileTypes=["csv"]
                ),
                "CSVLoader": build_file_path_template(
                    suffixes=[".csv"], fileTypes=["csv"]
                ),
                "UnstructuredEmailLoader": build_file_path_template(
                    suffixes=[".eml"], fileTypes=["eml"]
                ),
                "EverNoteLoader": build_file_path_template(
                    suffixes=[".xml"], fileTypes=["xml"]
                ),
                "FacebookChatLoader": build_file_path_template(
                    suffixes=[".json"], fileTypes=["json"]
                ),
                "GutenbergLoader": build_file_path_template(
                    suffixes=[".txt"], fileTypes=["txt"]
                ),
                "BSHTMLLoader": build_file_path_template(
                    suffixes=[".html"], fileTypes=["html"]
                ),
                "UnstructuredHTMLLoader": build_file_path_template(
                    suffixes=[".html"], fileTypes=["html"]
                ),
                "UnstructuredImageLoader": build_file_path_template(
                    suffixes=[".jpg", ".jpeg", ".png", ".gif", ".bmp"],
                    fileTypes=["jpg", "jpeg", "png", "gif", "bmp"],
                ),
                "UnstructuredMarkdownLoader": build_file_path_template(
                    suffixes=[".md"], fileTypes=["md"]
                ),
                "PyPDFLoader": build_file_path_template(
                    suffixes=[".pdf"], fileTypes=["pdf"]
                ),
                "UnstructuredPowerPointLoader": build_file_path_template(
                    suffixes=[".pptx", ".ppt"], fileTypes=["pptx", "ppt"]
                ),
                "SRTLoader": build_file_path_template(
                    suffixes=[".srt"], fileTypes=["srt"]
                ),
                "TelegramChatLoader": build_file_path_template(
                    suffixes=[".json"], fileTypes=["json"]
                ),
                "TextLoader": build_file_path_template(
                    suffixes=[".txt"], fileTypes=["txt"]
                ),
                "UnstructuredWordDocumentLoader": build_file_path_template(
                    suffixes=[".docx", ".doc"], fileTypes=["docx", "doc"]
                ),
                "SlackDirectoryLoader": build_file_path_template(
                    suffixes=[".zip"], fileTypes=["zip"]
                ),
            }

            if name in file_path_templates:
                signature["template"]["file_path"] = file_path_templates[name]
            elif name in {
                "WebBaseLoader",
                "AZLyricsLoader",
                "CollegeConfidentialLoader",
                "HNLoader",
                "IFixitLoader",
                "IMSDbLoader",
            }:
                signature["template"]["web_path"] = {
                    "type": "str",
                    "required": True,
                    "show": True,
                    "name": "web_path",
                    "value": "",
                    "display_name": "Web Page",
                }
            elif name in {"GitbookLoader"}:
                signature["template"]["web_page"] = {
                    "type": "str",
                    "required": True,
                    "show": True,
                    "name": "web_page",
                    "value": "",
                    "display_name": "Web Page",
                }
            elif name in {"ReadTheDocsLoader", "NotionDirectoryLoader"}:
                signature["template"]["path"] = {
                    "type": "str",
                    "required": True,
                    "show": True,
                    "name": "path",
                    "value": "",
                    "display_name": "Web Page",
                }

            return signature
        except ValueError as exc:
            raise ValueError(f"Documment Loader {name} not found") from exc
        except AttributeError as exc:
            logger.error(f"Documment Loader {name} not loaded: {exc}")
            return None

    def to_list(self) -> List[str]:
        return [
            documentloader.__name__
            for documentloader in self.type_to_loader_dict.values()
            if documentloader.__name__ in settings.documentloaders or settings.dev
        ]


documentloader_creator = DocumentLoaderCreator()
